// Copyright 2008-2018 Yolo Technologies, Inc. All Rights Reserved. https://www.comblockengine.com


#ifndef KBE_THREADPOOL_H
#define KBE_THREADPOOL_H

#include "common/common.h"
#include "common/tasks.h"
#include "helper/debug_helper.h"
#include "thread/threadtask.h"
// windows include	
#if KBE_PLATFORM == PLATFORM_WIN32
#include <windows.h>          // for HANDLE
#include <process.h>          // for _beginthread()	
#include "helper/crashhandler.h"
#else
// linux include
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <fcntl.h>
#include <sys/epoll.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pthread.h>	
#endif
	
namespace KBEngine{ namespace thread{

// ̳߳ػ̴߳Ŀڷæ״̬
#define THREAD_BUSY_SIZE 32

/*
	̳߳ص̻߳
*/
class ThreadPool;
class TPThread
{
public:
	friend class ThreadPool;

	// ߳״̬ -1δ, 0˯ߣ 1æ
	enum THREAD_STATE
	{
		THREAD_STATE_STOP = -1,
		THREAD_STATE_SLEEP = 0,
		THREAD_STATE_BUSY = 1,
		THREAD_STATE_END = 2,
		THREAD_STATE_PENDING = 3
	};

public:
	TPThread(ThreadPool* threadPool, int threadWaitSecond = 0):
	threadWaitSecond_(threadWaitSecond), 
	currTask_(NULL), 
	threadPool_(threadPool)
	{
		state_ = THREAD_STATE_SLEEP;
		initCond();
		initMutex();
	}
		
	virtual ~TPThread()
	{
		deleteCond();
		deleteMutex();

		DEBUG_MSG(fmt::format("TPThread::~TPThread(): {}\n", (void*)this));
	}
	
	virtual void onStart(){}
	virtual void onEnd(){}

	virtual void onProcessTaskStart(TPTask* pTask) {}
	virtual void processTask(TPTask* pTask){ pTask->process(); }
	virtual void onProcessTaskEnd(TPTask* pTask) {}

	INLINE THREAD_ID id(void) const;
	
	INLINE void id(THREAD_ID tidp);
	
	/**
		һ̣߳ Լ̰߳
	*/
	THREAD_ID createThread(void);
	
	virtual void initCond(void)
	{
		THREAD_SINGNAL_INIT(cond_);
	}

	virtual void initMutex(void)
	{
		THREAD_MUTEX_INIT(mutex_);	
	}

	virtual void deleteCond(void)
	{
		THREAD_SINGNAL_DELETE(cond_);
	}
	
	virtual void deleteMutex(void)
	{
		THREAD_MUTEX_DELETE(mutex_);
	}

	virtual void lock(void)
	{
		THREAD_MUTEX_LOCK(mutex_); 
	}
	
	virtual void unlock(void)
	{
		THREAD_MUTEX_UNLOCK(mutex_); 
	}	

	virtual TPTask* tryGetTask(void);
	
	/**
		ź
	*/
	int sendCondSignal(void)
	{
#if KBE_PLATFORM == PLATFORM_WIN32
		return THREAD_SINGNAL_SET(cond_);
#else
REATTEMPT:

		lock();

		if (state_ == THREAD_STATE_PENDING)
		{       
			unlock();
			goto REATTEMPT;
		}

		int ret = THREAD_SINGNAL_SET(cond_);
		unlock();
		return ret;
#endif
	}
	
	/**
		֪߳ͨ ȴź
	*/
	bool onWaitCondSignal(void);
	
	bool join(void);

	/**
		ȡ߳Ҫ
	*/
	INLINE TPTask* task(void) const;

	/**
		ñ߳Ҫ
	*/
	INLINE void task(TPTask* tpt);

	INLINE int state(void) const;
	
	/**
		߳ҪѾ Ǿɾ
	*/
	void onTaskCompleted(void);

#if KBE_PLATFORM == PLATFORM_WIN32
	static unsigned __stdcall threadFunc(void *arg);
#else	
	static void* threadFunc(void* arg);
#endif

	/**
		ñ߳Ҫ
	*/
	INLINE ThreadPool* threadPool();

	/**
		̹߳״̬
		Ҫṩwatcherʹ
	*/
	virtual std::string printWorkState()
	{
		char buf[128];
		lock();
		sprintf(buf, "%p,%u", currTask_, done_tasks_);
		unlock();
		return buf;
	}

	/**
		߳һδı䵽״ִ̬е
	*/
	void reset_done_tasks(){ done_tasks_ = 0; }
	void inc_done_tasks(){ ++done_tasks_; }

protected:
	THREAD_SINGNAL cond_;			// ߳ź
	THREAD_MUTEX mutex_;			// ̻߳
	int threadWaitSecond_;			// ߳̿״̬߳˳, С0Ϊ߳(뵥λ)
	TPTask * currTask_;				// ̵߳ĵǰִе
	THREAD_ID tidp_;				// ̵߳ID
	ThreadPool* threadPool_;		// ָ̳߳
	THREAD_STATE state_;			// ߳״̬: -1δ, 0˯, 1æ
	uint32 done_tasks_;				// ߳һδı䵽״ִ̬е
};


class ThreadPool
{
public:		
	
	ThreadPool();
	virtual ~ThreadPool();
	
	void finalise();

	virtual void onMainThreadTick();
	
	bool hasThread(TPThread* pTPThread);

	/**
		ȡǰ̳߳߳״̬(ṩwatch)
	*/
	std::string printThreadWorks();

	/**
		ȡǰ߳
	*/	
	INLINE uint32 currentThreadCount(void) const;
	
	/**
		ȡǰ߳
	*/		
	INLINE uint32 currentFreeThreadCount(void) const;
	
	/**
		̳߳
		@param inewThreadCount			: ϵͳæʱ̳߳ػộ߳ʱ
		@param inormalMaxThreadCount	: ̳߳ػһֱô߳
		@param imaxThreadCount			: ֻ̳߳ô߳
	*/
	bool createThreadPool(uint32 inewThreadCount, 
			uint32 inormalMaxThreadCount, uint32 imaxThreadCount);
	
	/**
		̳߳һ
	*/		
	bool addTask(TPTask* tptask);
	bool _addTask(TPTask* tptask);
	INLINE bool addBackgroundTask(TPTask* tptask){ return addTask(tptask); }
	INLINE bool pushTask(TPTask* tptask){ return addTask(tptask); }

	/**
		߳Ƿ񵽴
	*/
	INLINE bool isThreadCountMax(void) const;
	
	/**
		̳߳Ƿڷæ״̬
		δǷǳ   ˵̺ܷ߳æ
	*/
	INLINE bool isBusy(void) const;
	
	/** 
		̳߳ǷѾʼ 
	*/
	INLINE bool isInitialize(void) const;

	/**
		ǷѾ
	*/
	INLINE bool isDestroyed() const;

	/**
		ǷѾ
	*/
	INLINE void destroy();

	/** 
		û
	*/
	INLINE uint32 bufferTaskSize() const;

	/** 
		û
	*/
	INLINE std::queue<thread::TPTask*>& bufferedTaskList();

	/** 
		
	*/
	INLINE void lockBufferedTaskList();
	INLINE void unlockBufferedTaskList();

	/** 
		Ѿɵ
	*/
	INLINE uint32 finiTaskSize() const;

	virtual std::string name() const { return "ThreadPool"; }

public:
	static int timeout;

	/**
		һ̳߳߳
	*/
	virtual TPThread* createThread(int threadWaitSecond = ThreadPool::timeout, bool threadStartsImmediately = true);

	/**
		ĳ񱣴浽δб
	*/
	void bufferTask(TPTask* tptask);

	/**
		δбȡһ бɾ
	*/
	TPTask* popbufferTask(void);

	/**
		ƶһ̵߳б
	*/
	bool addFreeThread(TPThread* tptd);
	
	/**
		ƶһ̵߳æб
	*/	
	bool addBusyThread(TPThread* tptd);
	
	/**
		һѾɵб
	*/	
	void addFiniTask(TPTask* tptask);
	
	/**
		ɾһ(ʱ)߳
	*/	
	bool removeHangThread(TPThread* tptd);

	bool initializeWatcher();

protected:
	bool isInitialize_;												// ̳߳Ƿ񱻳ʼ
	
	std::queue<TPTask*> bufferedTaskList_;							// ϵͳڷæʱδб
	std::list<TPTask*> finiTaskList_;								// Ѿɵб
	size_t finiTaskList_count_;

	THREAD_MUTEX bufferedTaskList_mutex_;							// bufferTaskList
	THREAD_MUTEX threadStateList_mutex_;							// bufferTaskList and freeThreadList_
	THREAD_MUTEX finiTaskList_mutex_;								// finiTaskList
	
	std::list<TPThread*> busyThreadList_;							// æ߳б
	std::list<TPThread*> freeThreadList_;							// õ߳б
	std::list<TPThread*> allThreadList_;							// е߳б

	uint32 maxThreadCount_;											// ߳
	uint32 extraNewAddThreadCount_;									// normalThreadCount_㹻ʹ´ô߳
	uint32 currentThreadCount_;										// ǰ߳
	uint32 currentFreeThreadCount_;									// ǰõ߳
	uint32 normalThreadCount_;										// ׼״̬µ߳ ĬһͿô߳
																	// ̲߳㹻´һЩ̣߳ ܹmaxThreadNum.

	bool isDestroyed_;
};

}


}

#ifdef CODE_INLINE
#include "threadpool.inl"
#endif
#endif // KBE_THREADPOOL_H
